#include "DcsInterface.h"
#include "ConfigurationInterface.h"
#include "SctData/DcsData.h"
#include "Sct/SctNames.h"
#include "Sct/IoExceptions.h"
#include "Sct/LogicErrors.h"
#include "sctConf/configipc.h"

#include <is/isinfoany.h>
#include <iostream>
#include <memory>

namespace SctAnalysis{
  DcsInterface::DcsInterface() {}

  DcsInterface::DcsInterface(boost::shared_ptr<ConfigurationInterface> config)
    : m_config(config), m_call_count(0), m_val_count(0), m_error_count(0)
  {
  }

  DcsInterface::~DcsInterface(){
  }

  const ConfigurationInterface& DcsInterface::getConfigurationInterface() const {
      if (!m_config.get()) throw Sct::LogicError("No ConfigurationInterface available", __FILE__, __LINE__);
      return *m_config;
  }
  
  void DcsInterface::printStatus(std::ostringstream& os) const {
    std::cout << "DcsInterface::printStatus " << std::endl;
    os << "DcsInterface: Returned " << m_call_count << " DcsData objects"
       << " with total of " << m_val_count << " values." << std::endl;
    if (m_error_count) {
      os << "DcsInterface: Recorded " << m_error_count << " errors." << std::endl;
    }
    if (m_config.get()){
      try{
	m_config->print(os);
      }catch(std::exception& e){
	os << "DcsInterface has no configuration information" << std::endl;
      }
    }else{
      os << "DcsInterface has no configuration information" << std::endl;
    }
  }

  boost::shared_ptr<SctData::DcsData> DcsInterface::getData(const std::string& modulename) const{
    //std::cout << "DcsInterface::getData " <<  modulename << std::endl;
    ++ m_call_count;
    using SctData::DcsData;
    shared_ptr<DcsData> dcsdata;

    if (!m_config.get() || m_config->empty()){
       std::cerr << __FILE__ << ":" << __LINE__ 
	         << " No DCS mapping information available ... " 
		 << std::endl;
       return dcsdata;
    }
    
    string id = m_config->getDcsId(modulename);
    if (id=="") return dcsdata;  // no configuration mapping available

    string regexp;
    regexp = ".*"; regexp += id; regexp += ".*";
    
    ISInfoIterator it( Sct::SctNames::getPartition(), Sct::SctNames::getDcsServerName().c_str(), regexp.c_str() );
    
    while ( it() ){
      ++ m_val_count;
      ISInfoAny isa;
      it.value(isa);
      string name=it.name();
      
      //std::cout << "DcsInterface:: iterator " << it.name() << std::endl;

      if (isa.countAttributes()!=2) {
	std::cerr << "Error : DcsInterface:" << __LINE__
		  << " Received IS information `" 
		  << name << "' with " << isa.countAttributes() 
		  << " attributes. (Expected 2)" << std::endl;
	++ m_error_count;
	continue;
      }
      int dot1=name.rfind('.');
      if (dot1==string::npos || dot1<2){

      }
      int dot2=name.rfind('.',dot1-1);
      if (dot2==string::npos){
            std::cerr << "Error : DcsInterface:" << __LINE__ 
		      << " Received IS information `"
		      << name << std::endl; 
    	    ++m_error_count;
	    continue;
      } 
      std::string parname(name, dot2+1, dot1-dot2-1);
      if (parname==""){
            std::cerr << "Error: DcsInterface" << __LINE__
		      << " Received IS information `"
		      << name << "' parname = `" << parname 
		      << "' " << __FILE__ << ":" << __LINE__ << std::endl; 
	    ++m_error_count;
	    continue;
      }
      
      try {
	OWLTime value_time;
	isa>>value_time;

	float val;
	switch(isa.getAttributeType()){
	case ISType::S8:
	case ISType::U8:
	case ISType::S16:
	  {
	    short temp;
	    isa >> temp;
	    val=temp;
	    break;
	  }
	case ISType::U16:
	  {
	    unsigned short temp;
	    isa >> temp;
	    val=temp;
	    break;
	  }
	case ISType::Float:
	  {
	    isa >> val;
	    break;
	  }
	case ISType::Double:
	{
	  double temp;
	  isa >> temp;
	  val=temp;
	  break;
	}
	default:
	  {
	    std::cerr << __FILE__ << ":" << __LINE__ << "Unexpected attribute type:" 
		      << isa.getAttributeType() << std::endl; 
	    continue;
	  }
	}
	// initialise if necessary
	if (!dcsdata.get()) dcsdata = shared_ptr<DcsData>( new DcsData() );
	dcsdata->setParameter(parname, val, value_time);
	//std::cout << "Parname = `" << parname << "' = " << val << std::endl;
      }catch(std::exception& e){
	std::cerr << e.what() << " " << __FILE__ << ":" << __LINE__ << std::endl;
	++m_error_count;
	break;
      }
    } // end of loop over parameters

    if (!dcsdata.get()){
      std::cerr << "Failed to match patterm `" << regexp << "'" << " in server `"
		<< Sct::SctNames::getDcsServerName() << "'" << std::endl;
    }
    return dcsdata;
  }
}

