#include "TxCurrentConfigUpdater.h"
#include "ConfigUpdaterManager.h"
#include "../SctApiAccessException.h"
#include "../SctApiCall.h"

#include "Sct_SctApi/Sct_SctApi.hh"
#include "Sct/SctParameters.h"
#include "Sct/SctNames.h"
#include "SctData/TxCurrentTestResult.h"
#include "CommonWithDsp/ABCD/ABCDscans.h"
#include "SctApi/utility.h"

using namespace Sct;
using namespace SctData;

namespace SctCalibrationController {

bool TxCurrentConfigUpdater::inMap = ConfigUpdaterManager::instance().setUpdater("SctData::TxCurrentTestResult", shared_ptr<ConfigUpdater> (new TxCurrentConfigUpdater()));

  void TxCurrentConfigUpdater::complainIf(const int threshold,
                                          const unsigned int linkNumber,
                                          const unsigned long mid,
                                          const std::string & name,
                                          const int otherMin,
                                          const int otherMax) {
    if ((threshold>otherMax && TxCurrentTestResult::StreamResult::isGoodPhysicalTxCurrent(otherMax)) || (threshold<otherMin && TxCurrentTestResult::StreamResult::isGoodPhysicalTxCurrent(otherMin))) {
      unsigned partition, crate, rod, channel;
      SctApi::Utility::getpcrc(mid, partition, crate, rod, channel);
      std::ostringstream os;
      os << "Warning: while setting the TxCurrent for module " << mid << " named " << name << " in prcc " << partition << "/" << crate << "/" << rod << "/" << channel << " at line " << __LINE__<< " of " << __FILE__ << " it was thought best to use the data returned from link " << linkNumber<< " as it appeared slightly more reliable.  This link indicated that a TxCurrent setting of " << threshold << " would be sensible.  The problem is that the value of TxCurrent we have just chosen to use is flaggeed as *unacceptable* by the data from the other link which expected a value lying in the range ["<<otherMin<<", "<<otherMax<<"].  So either link "<<(1-linkNumber)<< " is not working properly even though I thought it was fine.  You had better check the cause of this inconsistency.";
      SctNames::Mrs() << "CC_UPDATE" << MRS_WARNING << MRS_TEXT(os.str()) << ENDM;
    };
  };

  void TxCurrentConfigUpdater::updateUsing(const int threshold,
                                           const unsigned int linkNumber,
                                           Sct_SctApi_T_SctApi& api,
                                           const unsigned long mid,
                                           const std::string & name) {
    unsigned partition, crate, rod, channel;
    SctApi::Utility::getpcrc(mid, partition, crate, rod, channel);

    if (TxCurrentTestResult::StreamResult::isGoodPhysicalTxCurrent(threshold)) {  
      APICALL(&api, modifyBOCParam(&st, partition, crate, rod, channel, ST_TX_CURRENT, threshold ), "TxCurrentConfigUpdater failed to set delay") ;
      std::ostringstream os;
      os << "Setting Register ST_TX_CURRENT=" << ST_TX_CURRENT << " based on data from link " << linkNumber << " of module " << mid << " named " << name << " in prcc " << partition << "/" << crate << "/" << rod << "/" << channel << " to " << threshold;
      SctNames::Mrs() << "CC_UPDATE" << MRS_INFORMATION << MRS_TEXT(os.str()) << ENDM;
    } else {
      std::ostringstream os;
      os << "Warning: was unable to set the TxCurrent based on data from link "<<linkNumber<<" of module " << mid << " in prcc " << partition << "/" << crate << "/" << rod << "/" << channel << " as the test was not able to determine a sensible value for it! ["<<threshold<<"]";
      SctNames::Mrs() << "CC_UPDATE" << MRS_WARNING << MRS_TEXT(os.str()) << ENDM;
    };
  };

void TxCurrentConfigUpdater::update(const TestResult& testresult, Sct_SctApi_T_SctApi& api) const {
    const TxCurrentTestResult& t = dynamic_cast<const TxCurrentTestResult&> (testresult);

    unsigned long mid = getMID(t, api);
    SctNames::Mrs() << "CC_UPDATE" << MRS_TEXT("CalibrationController updating TxCurrent") << MRS_PARAM<long>("ModuleID", mid) << MRS_DIAGNOSTIC << ENDM;

    const TxCurrentTestResult::StreamResult & sr0 = t.getStreamResult(0);
    const TxCurrentTestResult::StreamResult & sr1 = t.getStreamResult(1);
    
    const int link0ThreshMin = sr0.minErrorFreeTxCurrent();
    const int link1ThreshMin = sr1.minErrorFreeTxCurrent();
    const int link0ThreshMax = sr0.maxErrorFreeTxCurrent();
    const int link1ThreshMax = sr1.maxErrorFreeTxCurrent();
    const int link0ThreshBest = sr0.bestTxCurrent();
    const int link1ThreshBest = sr1.bestTxCurrent();
    const int link0Width = link0ThreshMax - link0ThreshMin;
    const int link1Width = link1ThreshMax - link1ThreshMin;
    const bool link0Invalid = sr0.isInvalid(); 
    const bool link1Invalid = sr1.isInvalid();
 
    const std::string & name = t.getModuleName();

    if (link0Invalid && link1Invalid) {
      // both bad
    } else if (link0Invalid) {
      // link 1 ok
      this->updateUsing(link1ThreshBest, 1, api, mid, name);
    } else if (link1Invalid) {
      // link 0 ok
      this->updateUsing(link0ThreshBest, 0, api, mid, name);
    } else {
      // Both links ok.
      // For setting tx current we want (where possible) to be INSENSITVE to problems that there might be with the ability of the modules to send, so if one of the streams is worse than the other it is nothing to do with us.  We therefore set on the basis of the BEST (widest) window we see.
      if (link0Width>link1Width) {
        this->updateUsing(link0ThreshBest, 0, api, mid, name);
        this-> complainIf(link0ThreshBest, 0,      mid, name, link1ThreshMin, link1ThreshMax);
      } else {
        this->updateUsing(link1ThreshBest, 1, api, mid, name);
        this-> complainIf(link1ThreshBest, 1,      mid, name, link0ThreshMin, link0ThreshMax);
      };
    };
}
    
}
